package me.totalfreedom.totalfreedommod.sql;

import me.totalfreedom.totalfreedommod.FreedomService;
import me.totalfreedom.totalfreedommod.admin.Admin;
import me.totalfreedom.totalfreedommod.banning.Ban;
import me.totalfreedom.totalfreedommod.player.PlayerData;
import me.totalfreedom.totalfreedommod.util.FLog;
import me.totalfreedom.totalfreedommod.util.FUtil;

import java.sql.*;
import java.text.MessageFormat;
import java.util.UUID;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.CompletionException;

public class SQLite extends FreedomService
{
    public static final String TABLE_CHECK_FAILURE = "Failed to check tables on database: ";
    private final String FILE_NAME = "database.db";

    private Connection connection;

    @Override
    public void onStart()
    {
        connect();
        checkTables().whenComplete((result, throwable) ->
        {
            if (throwable != null)
            {
                FLog.severe(TABLE_CHECK_FAILURE + throwable.getMessage());
            }
        });
    }

    @Override
    public void onStop()
    {
        disconnect();
    }

    public void connect()
    {
        try
        {
            connection = DriverManager.getConnection("jdbc:sqlite:" + plugin.getDataFolder() + "/" + FILE_NAME);
            FLog.info("Successfully connected to the database.");
        } catch (SQLException e)
        {
            FLog.severe("Failed to connect to the database: " + e.getMessage());
        }
    }

    public void disconnect()
    {
        try
        {
            if (connection != null)
            {
                connection.close();
            }
        } catch (SQLException e)
        {
            FLog.severe("Failed to disconnect from the database: " + e.getMessage());
        }
    }

    public CompletableFuture<Void> checkTables()
    {
        return CompletableFuture.supplyAsync(() ->
        {
            try
            {
                return connection.getMetaData();
            } catch (SQLException e)
            {
                throw new CompletionException(TABLE_CHECK_FAILURE, e);
            }
        }).thenAcceptAsync(meta ->
        {
            futureBansTable(meta);

            futureAdminsTable(meta);

            futurePlayersTable(meta);
        });
    }

    private void futurePlayersTable(DatabaseMetaData meta)
    {
        tableNotExists(meta, "players").whenComplete((result, throwable) ->
        {
            if (throwable != null)
            {
                FLog.severe(throwable.getMessage());
                return;
            }

            if (result)
            {
                createPlayersTable().whenComplete((result2, throwable2) ->
                {
                    if (throwable2 != null)
                    {
                        FLog.severe(throwable2.getMessage());
                    }
                });
            }
        });
    }

    private void futureAdminsTable(DatabaseMetaData meta)
    {
        tableNotExists(meta, "admins").whenComplete((result, throwable) ->
        {
            if (throwable != null)
            {
                FLog.severe(throwable.getMessage());
                return;
            }

            if (result)
            {
                createAdminsTable().whenComplete((result2, throwable2) ->
                {
                    if (throwable2 != null)
                    {
                        FLog.severe(throwable2.getMessage());
                    }
                });
            }
        });
    }

    private void futureBansTable(DatabaseMetaData meta)
    {
        tableNotExists(meta, "bans").whenComplete((result, throwable) ->
        {
            if (throwable != null)
            {
                FLog.severe(throwable.getMessage());
                return;
            }

            if (Boolean.TRUE.equals(result)) // using primitive for some reason... ask SonarLint.
            {
                createBanTable().whenComplete((result2, throwable2) ->
                {
                    if (throwable2 != null)
                    {
                        FLog.severe(throwable2.getMessage());
                    }
                });
            }
        });
    }

    private CompletableFuture<Void> createPlayersTable()
    {
        return CompletableFuture.supplyAsync(() ->
        {
            try (PreparedStatement statement = connection.prepareStatement("CREATE TABLE `players` (`uuid` VARCHAR NOT NULL, `ips` VARCHAR NOT NULL, `notes` VARCHAR, `tag` VARCHAR, `discord_id` VARCHAR, `master_builder` BOOLEAN NOT NULL, `ride_mode` VARCHAR NOT NULL, `coins` INT, `items` VARCHAR, `total_votes` INT NOT NULL, `display_discord` BOOLEAN NOT NULL, `login_message` VARCHAR, `inspect` BOOLEAN NOT NULL);"))
            {
                statement.executeUpdate();
            } catch (SQLException e)
            {
                throw new CompletionException("Failed to create the players table.", e);
            }
            return null;
        });
    }

    private CompletableFuture<Void> createAdminsTable()
    {
        return CompletableFuture.supplyAsync(() ->
        {
            try (PreparedStatement statement = connection.prepareStatement("CREATE TABLE `admins` (`uuid` VARCHAR NOT NULL, `ips` VARCHAR NOT NULL, `rank` VARCHAR NOT NULL, `active` BOOLEAN NOT NULL, `last_login` LONG NOT NULL, `command_spy` BOOLEAN NOT NULL, `potion_spy` BOOLEAN NOT NULL, `ac_format` VARCHAR);"))
            {
                statement.executeUpdate();
            } catch (SQLException e)
            {
                throw new CompletionException("Failed to create the admins table.", e);
            }
            return null;
        });
    }

    private CompletableFuture<Void> createBanTable()
    {
        return CompletableFuture.supplyAsync(() ->
        {
            try (PreparedStatement statement = connection.prepareStatement("CREATE TABLE `bans` (`name` VARCHAR, `uuid` VARCHAR, `ips` VARCHAR, `by` VARCHAR NOT NULL, `at` LONG NOT NULL, `expires` LONG, `reason` VARCHAR);"))
            {
                statement.executeUpdate();
            } catch (SQLException e)
            {
                throw new CompletionException("Failed to create the bans table.", e);
            }
            return null;
        });
    }

    public CompletableFuture<Void> truncate(String table)
    {
        return CompletableFuture.supplyAsync(() ->
        {
            try (PreparedStatement statement = connection.prepareStatement("DELETE FROM ?"))
            {
                statement.setString(1, table);
                statement.executeUpdate();
            } catch (SQLException e)
            {
                FLog.severe("Failed to truncate " + table + ": " + e.getMessage());
            }
            return null;
        });
    }

    public CompletableFuture<ResultSetProvider> getBanList()
    {
        return CompletableFuture.supplyAsync(() ->
        {
            ResultSetProvider provider;
            try (PreparedStatement statement = connection.prepareStatement("SELECT * FROM bans"))
            {
                try (ResultSet result = statement.executeQuery())
                {
                    provider = ResultSetProvider.provideAllRows(result);
                }
            } catch (SQLException e)
            {
                throw new CompletionException("Failed to get ban list.", e);
            }
            return provider;
        });
    }

    public CompletableFuture<ResultSetProvider> getAdminList()
    {
        return CompletableFuture.supplyAsync(() ->
        {
            ResultSetProvider provider;
            try (PreparedStatement statement = connection.prepareStatement("SELECT * FROM bans"))
            {
                try (ResultSet result = statement.executeQuery())
                {
                    provider = ResultSetProvider.provideAllRows(result);
                }
            } catch (SQLException ex)
            {
                throw new CompletionException("Failed to get admin list.", ex);
            }
            return provider;
        });

    }

    public CompletableFuture<Void> setAdminValue(Admin admin, String key, Object value)
    {
        return CompletableFuture.supplyAsync(() ->
        {
            try
            {
                Object[] data = {key, admin.getUuid()};
                PreparedStatement statement = connection.prepareStatement(MessageFormat.format("UPDATE admins SET {0}=? WHERE uuid=''{1}''", data));
                setUnknownType(statement, 1, value).whenComplete((result, throwable) ->
                {
                    if (throwable != null)
                    {
                        FLog.severe(throwable.getMessage());
                    }

                    try
                    {
                        result.executeUpdate();
                    } catch (SQLException ex)
                    {
                        throw new CompletionException("Failed to update admin value.", ex);
                    }
                });
            } catch (SQLException e)
            {
                throw new CompletionException("Failed to update admin value.", e);
            }
            return null;
        });
    }

    public CompletableFuture<Void> setPlayerValue(PlayerData player, String key, Object value)
    {
        return CompletableFuture.supplyAsync(() ->
        {
            try
            {
                Object[] data = {key, player.getUuid()};
                PreparedStatement statement = connection.prepareStatement(MessageFormat.format("UPDATE players SET {0}=? WHERE uuid=''{1}''", data));
                setUnknownType(statement, 1, value).whenComplete((result, throwable) ->
                {
                    if (throwable != null)
                    {
                        FLog.severe(throwable.getMessage());
                    }

                    try
                    {
                        result.executeUpdate();
                    } catch (SQLException ex)
                    {
                        throw new CompletionException("Failed to update player value.", ex);
                    }
                });
            } catch (SQLException e)
            {
                throw new CompletionException("Failed to update player value.", e);
            }
            return null;
        });
    }

    public CompletableFuture<PreparedStatement> setUnknownType(PreparedStatement statement, int index, Object value)
    {
        return CompletableFuture.supplyAsync(() ->
        {
            try
            {
                if (value == null)
                {
                    statement.setString(index, null);
                } else if (value.getClass().equals(String.class))
                {
                    String v = (String) value;
                    statement.setString(index, v);
                } else if (value.getClass().equals(Integer.class))
                {
                    int v = (int) value;
                    statement.setInt(index, v);
                } else if (value.getClass().equals(Boolean.class))
                {
                    boolean v = (boolean) value;
                    statement.setBoolean(index, v);
                } else if (value.getClass().equals(Long.class))
                {
                    long v = (long) value;
                    statement.setLong(index, v);
                }
                return statement;
            } catch (SQLException ex)
            {
                throw new CompletionException("Failed to set unknown type.", ex);
            }
        });
    }

    public Object getValue(ResultSetProvider provider, String key, Object value)
    {
        Object result = null;
        if (value instanceof String)
        {
            result = provider.getString(key);
        } else if (value instanceof Integer)
        {
            result = provider.getInt(key);
        } else if (value instanceof Boolean)
        {
            result = provider.getBoolean(key);
        } else if (value instanceof Long)
        {
            result = provider.getLong(key);
        }
        return result;
    }

    public CompletableFuture<Void> addAdmin(Admin admin)
    {
        return CompletableFuture.supplyAsync(() ->
        {
            try (PreparedStatement statement = connection.prepareStatement("INSERT INTO admins VALUES (?, ?, ?, ?, ?, ?, ?, ?)"))
            {
                statement.setString(1, admin.getUuid().toString());
                statement.setString(2, FUtil.listToString(admin.getIps()));
                statement.setString(3, admin.getRank().toString());
                statement.setBoolean(4, admin.isActive());
                statement.setLong(5, admin.getLastLogin().getTime());
                statement.setBoolean(6, admin.getCommandSpy());
                statement.setBoolean(7, admin.getPotionSpy());
                statement.setString(8, admin.getAcFormat());
                statement.executeUpdate();
            } catch (SQLException e)
            {
                throw new CompletionException("Failed to add admin.", e);
            }
            return null;
        });

    }

    public CompletableFuture<Void> addPlayer(PlayerData player)
    {
        return CompletableFuture.supplyAsync(() ->
        {
            try (PreparedStatement statement = connection.prepareStatement("INSERT INTO players VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"))
            {
                statement.setString(1, player.getUuid().toString());
                statement.setString(2, FUtil.listToString(player.getIps()));
                statement.setString(3, FUtil.listToString(player.getNotes()));
                statement.setString(4, FUtil.miniMessage(player.getTag()));
                statement.setString(5, player.getDiscordID());
                statement.setBoolean(6, player.isMasterBuilder());
                statement.setString(7, player.getRideMode().name());
                statement.setInt(8, player.getCoins());
                statement.setString(9, FUtil.listToString(player.getItems()));
                statement.setInt(10, player.getTotalVotes());
                statement.setBoolean(11, player.doesDisplayDiscord());
                statement.setString(12, player.getLoginMessage());
                statement.setBoolean(13, player.hasInspection());
                statement.executeUpdate();
            } catch (SQLException e)
            {
                throw new CompletionException("Failed to add player.", e);
            }
            return null;
        });
    }

    public CompletableFuture<ResultSetProvider> getAdminByUuid(UUID uuid)
    {
        return CompletableFuture.supplyAsync(() ->
        {
            ResultSetProvider provider;
            try (PreparedStatement statement = connection.prepareStatement("SELECT * FROM admins WHERE uuid=?"))
            {
                statement.setString(1, uuid.toString());
                try (ResultSet set = statement.executeQuery())
                {
                    provider = ResultSetProvider.provideSpecificRow(set);
                }
            } catch (SQLException ex)
            {
                throw new CompletionException(ex);
            }
            return provider;
        });
    }

    public CompletableFuture<ResultSetProvider> getPlayerByUuid(UUID uuid)
    {
        return CompletableFuture.supplyAsync(() ->
        {
            ResultSetProvider provider;
            try (PreparedStatement statement = connection.prepareStatement("SELECT * FROM players WHERE uuid=?"))
            {
                statement.setString(1, uuid.toString());
                try (ResultSet set = statement.executeQuery())
                {
                    provider = ResultSetProvider.provideSpecificRow(set);
                }
            } catch (SQLException ex)
            {
                throw new CompletionException(ex);
            }

            return provider;
        });
    }

    public CompletableFuture<ResultSetProvider> getMasterBuilders()
    {
        return CompletableFuture.supplyAsync(() ->
        {
            ResultSetProvider provider;
            try (PreparedStatement statement = connection.prepareStatement("SELECT * FROM players WHERE master_builder=true"))
            {
                try (ResultSet result = statement.executeQuery())
                {
                    provider = ResultSetProvider.provideAllRows(result);
                }
            } catch (SQLException ex)
            {
                throw new CompletionException(ex);
            }
            return provider;
        });
    }

    public CompletableFuture<ResultSetProvider> getPlayerByIp(String ip)
    {
        return CompletableFuture.supplyAsync(() ->
        {
            ResultSetProvider provider;
            try (PreparedStatement statement = connection.prepareStatement("SELECT * FROM players WHERE ips LIKE %?%"))
            {
                statement.setString(1, ip);
                try (ResultSet set = statement.executeQuery())
                {
                    provider = ResultSetProvider.provideSpecificRow(set);
                }
            } catch (SQLException e)
            {
                throw new CompletionException(e);
            }

            return provider;
        });
    }

    public CompletableFuture<Void> removeAdmin(Admin admin)
    {
        return CompletableFuture.supplyAsync(() ->
        {
            try (PreparedStatement statement = connection.prepareStatement("DELETE FROM admins where name=?"))
            {
                statement.setString(1, admin.getName());
                statement.executeUpdate();
            } catch (SQLException e)
            {
                throw new CompletionException("Failed to remove admin.", e);
            }
            return null;
        });
    }


    public CompletableFuture<Void> addBan(Ban ban)
    {
        return CompletableFuture.supplyAsync(() ->
        {
            try (PreparedStatement statement = connection.prepareStatement("INSERT INTO bans VALUES (?, ?, ?, ?, ?, ?, ?)"))
            {
                statement.setString(1, ban.getUsername());
                String uuid = null;
                if (ban.hasUUID())
                {
                    uuid = ban.getUuid().toString();
                }
                statement.setString(2, uuid);
                statement.setString(3, FUtil.listToString(ban.getIps()));
                statement.setString(4, ban.getBy());
                statement.setLong(5, ban.getAt().getTime());
                statement.setLong(6, ban.getExpiryUnix());
                statement.setString(7, ban.getReason());
                statement.executeUpdate();
            } catch (SQLException e)
            {
                throw new CompletionException("Failed to add ban.", e);
            }
            return null;
        });
    }

    public CompletableFuture<Void> removeBan(Ban ban)
    {
        return CompletableFuture.supplyAsync(() ->
        {
            try
            {
                try (PreparedStatement statement = connection.prepareStatement("DELETE FROM bans WHERE name=?"))
                {
                    statement.setString(1, ban.getUsername());
                    statement.executeUpdate();
                }

                for (String ip : ban.getIps())
                {
                    try (PreparedStatement statement = connection.prepareStatement("DELETE FROM bans WHERE ips LIKE %?%"))
                    {
                        statement.setString(1, ip);
                        statement.executeUpdate();
                    }
                }
            } catch (SQLException e)
            {
                throw new CompletionException("Failed to remove ban.", e);
            }
            return null;
        });
    }

    // We've changed this to read tableNotExists because it's more accurate in context.
    public CompletableFuture<Boolean> tableNotExists(DatabaseMetaData meta, String name)
    {
        return CompletableFuture.supplyAsync(() ->
        {
            try
            {
                return !meta.getTables(null, null, name, null).next();
            } catch (SQLException ex)
            {
                throw new CompletionException(TABLE_CHECK_FAILURE, ex);
            }
        });
    }
}